@@ -54,6 +54,8 @@ Only the background color is significant."
 (defvar lv-wnd nil
   "Holds the current LV window.")
 
+(defvar display-line-numbers)
+
 (defun lv-window ()
   "Ensure that LV window is live and return it."
   (if (window-live-p lv-wnd)
@@ -72,6 +74,7 @@ Only the background color is significant."
           (setq window-size-fixed t)
           (setq mode-line-format nil)
           (setq cursor-type nil)
+          (setq display-line-numbers nil)
           (set-window-dedicated-p lv-wnd t)
           (set-window-parameter lv-wnd 'no-other-window t))
         (select-window ori)))))
